/*
 * $Id: RapiConnection.cs 287 2007-06-06 07:05:28Z lloydd $
 */

using System;
using System.IO;
using System.Net;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using Mono.Unix;
using Mono.GetOptions;
using NDesk.DBus;
using org.freedesktop.DBus;
using org.wimo.dccm;

namespace Dccm
{
   enum RapiState {
      HandShake,
      GettingInfo,
      Auth,
      ConnectionEstablished
   }
   
   public class RapiConnection
   {
      Device device;
      string rapi_ip;
      int rapi_port;
      bool device_added;
      NetworkStream stream;
      TcpClient device_sck;
      string ip_addr;
      RapiState state;
      RapiForwardServer fwd_srv;
      
      public Device Device {
         get {
            return device;
         }
      }
      
      public TcpClient DeviceSck {
         get {
            return device_sck;
         }
      }
      
      public RapiConnection (TcpClient device_sck, string rapi_ip, int rapi_port)
      {
         this.device_sck = device_sck;
         this.rapi_ip = rapi_ip;
         this.rapi_port = rapi_port;
         stream = this.device_sck.GetStream ();
         
         device = null;
         device_added = false;
         state = RapiState.HandShake;
         ip_addr = ((IPEndPoint) device_sck.Client.RemoteEndPoint).Address.ToString ();

         fwd_srv = new RapiForwardServer (device_sck, rapi_port);
         
         if (MainClass.Opt.Debug)
            Console.WriteLine ("RAPI client connection from {0}", ip_addr);
      }
      
      public void Start ()
      {
         new Thread (new ThreadStart (Handler)).Start ();
      }
      
      public void Handler ()
      {
         try {
            
            byte[] buf = BitConverter.GetBytes ((uint) 3);
            stream.Write (buf, 0, buf.Length);
            
            while (Process ());

            fwd_srv.Start ();
            
            device_sck.NoDelay = true;
            device_sck.ReceiveTimeout = 3000;
            
            while (TestConnection ())
               Thread.Sleep (8000);
         }
         catch (Exception e) {
            
            Console.Error.WriteLine ("Exception: {0}", e.Message);
         }

         fwd_srv.Stop ();
         
         if (device != null) {
            
            MainClass.Manager.Disconnected (device);
            
            if (device_added)
               MainClass.Manager.Remove (device);
         }
      }
      
      public bool SendPassword (string psw, uint key)
      {
         byte[] encoded = new byte [2 * (psw.Length + 1)];
         
         for (int i = 0; i < psw.Length; i++)
            encoded [i * 2] = (byte) psw [i];

         for (int i = 0; i < encoded.Length; i++)
            encoded [i] = (byte)(encoded [i] ^ key);
         
         byte[] len = BitConverter.GetBytes ((ushort) encoded.Length);
         
         stream.Write (len, 0, len.Length);
         stream.Write (encoded, 0, encoded.Length);

         len = new byte [2];

         if (stream.Read (len, 0, len.Length) != len.Length)
            return false;

         return len [0] == 1;
      }
      
      public bool TestConnection ()
      {
         byte[] pong = new byte [sizeof (uint)];
         byte[] ping = BitConverter.GetBytes ((uint) 1);

         if (MainClass.Opt.Debug)
            Console.WriteLine ("Sending ping...");
         
         stream.Write (ping, 0, ping.Length);

         if (MainClass.Opt.Debug)
            Console.WriteLine ("Receiving pong...");
         
         if (stream.Read (pong, 0, pong.Length) != pong.Length) {

            if (MainClass.Opt.Debug)
               Console.WriteLine ("Pong receive failed...");

            return false;
         }

         if (MainClass.Opt.Debug)
            Console.WriteLine ("Pong received...");
         
         return true;
      }
      
      public bool Process ()
      {
         byte[] tmp = null;
         byte[] buf = new byte [sizeof (uint)];
         uint req = 0;
         uint len = 0;
         
         if (stream.Read (buf, 0, buf.Length) != buf.Length)
            return false;
         
         req = BitConverter.ToUInt32 (buf, 0);
         
         if (MainClass.Opt.Debug)
            Console.WriteLine ("State: {0}", state);
         
         if (state == RapiState.HandShake) {
            
            if (MainClass.Opt.Debug)
               Console.WriteLine ("Received handshake request: {0}", req);
               
            switch (req) {
            case 4:
               state = RapiState.GettingInfo;
               break;
            case 6:
               buf = new byte [4 * sizeof (uint)];
               
               tmp = BitConverter.GetBytes ((uint) 7);
               tmp.CopyTo (buf, 0 * sizeof (uint));
               
               tmp = BitConverter.GetBytes ((uint) 8);
               tmp.CopyTo (buf, 1 * sizeof (uint));
               
               tmp = BitConverter.GetBytes ((uint) 4);
               tmp.CopyTo (buf, 2 * sizeof (uint));
               
               tmp = BitConverter.GetBytes ((uint) 5);
               tmp.CopyTo (buf, 3 * sizeof (uint));
               
               stream.Write (buf, 0, buf.Length);
               break;
            }
         }
            
         if (state == RapiState.GettingInfo) {

            if (MainClass.Opt.Debug)
               Console.WriteLine ("Receiving size of device info");
            
            buf = new byte [sizeof (uint)];
            
            if (stream.Read (buf, 0, buf.Length) != buf.Length)
               return false;
            
            len = BitConverter.ToUInt32 (buf, 0);
            
            if (MainClass.Opt.Debug)
               Console.WriteLine ("Receiving device info size: {0}", len);
            
            buf = new byte [len];

            if (stream.Read (buf, 0, buf.Length) != buf.Length)
               return false;

            if (MainClass.Opt.Debug)
               Console.WriteLine ("Received device info");
            
            if ((device = ProcessInfo (buf)) != null) {
            
               if (MainClass.Opt.Debug) {
                  
                  Console.WriteLine ("Device info parsed");
                  Console.WriteLine ("Guid: \"{0}\"", device.Guid);
                  Console.WriteLine ("OsVersion: {0}.{1}", device.OsVersionMajor, device.OsVersionMinor);
                  Console.WriteLine ("Name: \"{0}\"", device.Name);
                  Console.WriteLine ("PlatformName: \"{0}\"", device.PlatformName);
                  Console.WriteLine ("ModelName: \"{0}\"", device.ModelName);
                  Console.WriteLine ("DeviceIP: \"{0}\"", device.DeviceIP);
                  Console.WriteLine ("DeviceRapiIP: \"{0}\"", device.DeviceRapiIP);
                  Console.WriteLine ("DeviceRapiPort: \"{0}\"", device.DeviceRapiPort);
                  Console.WriteLine ("Version: 0x{0}", device.Version.ToString ("X08"));
                  Console.WriteLine ("CpuType: {0}", device.CpuType);
                  Console.WriteLine ("CurrentPartnerId: {0}", device.CurrentPartnerId);
                  Console.WriteLine ("Id: {0}", device.Id);
                  Console.WriteLine ("Key: {0}", device.Key);
                  Console.WriteLine ("PasswordProtected: {0}", device.PasswordProtected);
                  
                  if (device.PasswordProtected)
                     state = RapiState.Auth;
                  else
                     state = RapiState.ConnectionEstablished;

                  MainClass.Manager.Add (device);

                  if (MainClass.Opt.Debug)
                     Console.WriteLine ("Signaling connected");

                  MainClass.Manager.Connected (device);
               }
            } else
               Console.Error.WriteLine ("Failed to parse device info buffer");
         }

         if (state == RapiState.Auth) {

            if (MainClass.Opt.Debug)
               Console.WriteLine ("Signaling password request");
            
            device.PasswordProvided = false;
            
            MainClass.Manager.PasswordRequested (device);

            int cnt = 15*4;
            
            while (!device.PasswordProvided)
               if (cnt-- > 0)
                  Thread.Sleep (250);
               else
                  break;
            
            if (device.PasswordProvided) {
               
               if (MainClass.Opt.Debug)
                  Console.WriteLine ("Password provided");
               
               if (SendPassword (device.Password, device.Key)) {
                  
                  if (MainClass.Opt.Debug)
                     Console.WriteLine ("Password accepted");
            
                  Thread.Sleep (250); // gives some time to the device
                  
                  state = RapiState.ConnectionEstablished;
               } else {
                  
                  if (MainClass.Opt.Debug)
                     Console.WriteLine ("Password not accepted");
                  
                  MainClass.Manager.WrongPassword (device);
                  
                  return false;
               }
            } else {
               
               if (MainClass.Opt.Debug)
                  Console.WriteLine ("Password not provided");
               
               return false;
            }
         }
         
         if (state == RapiState.ConnectionEstablished) {

            if (MainClass.Opt.Debug)
               Console.WriteLine ("Signaling connection established");

            MainClass.Manager.ConnectionEstablished (device);
            
            return false;
         }

         return true;
      }
      
      public Device ProcessInfo (byte[] buf)
      {
         int p = 0;
         
         if (MainClass.Opt.Debug)
            WriteBuf ("Device info buffer", buf, buf.Length);

         string guid = 
            BitConverter.ToUInt32 (buf, 0).ToString ("X08") +
            "-" +
            BitConverter.ToUInt16 (buf, 4).ToString ("X04") +
            "-" +
            BitConverter.ToUInt16 (buf, 6).ToString ("X04") +
            "-" +
            buf [8].ToString ("X02") +
            buf [9].ToString ("X02") +
            "-" +
            buf [10].ToString ("X02") +
            buf [11].ToString ("X02") +
            buf [12].ToString ("X02") +
            buf [13].ToString ("X02") +
            buf [14].ToString ("X02") +
            buf [15].ToString ("X02");
         
         p = 16;
         uint os_major = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         uint os_minor = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         string name = GetUnicodeString (buf, ref p);
         uint version = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         uint cpu_type = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         // counter of flags?
         p += sizeof (uint);
         uint current_partner_id = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         uint id = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
         string platform_name = GetAsciiString (buf, ref p);
         string model_name = GetAsciiString (buf, ref p);
         p ++;
         
         Device dev = new Device ();
         dev.OsVersionMajor = os_major;
         dev.OsVersionMinor = os_minor;
         dev.Guid = guid;
         dev.Name = name;
         dev.Version = version;
         dev.CpuType = cpu_type;
         dev.CurrentPartnerId = current_partner_id;
         dev.Id = id;
         dev.PlatformName = platform_name;
         dev.ModelName = model_name;
         dev.DeviceIP = ip_addr;
         dev.DeviceRapiIP = rapi_ip;
         dev.DeviceRapiPort = (uint) rapi_port;
         dev.PasswordProtected = false;
         
         if ((p + sizeof (uint)) > buf.Length)
            return dev;
         
         uint comp_count = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);
 
         // ignore component versions for now
         p += (int) (comp_count * (sizeof (uint) * 2));
         
         if ((p + sizeof (uint)) > buf.Length)
            return dev;
         
         dev.Key = BitConverter.ToUInt32 (buf, p);
         p += sizeof (uint);

         // Key == 0xffffffff than device is locked (introduces by WM6)
         
         dev.PasswordProtected = dev.Key != 0 && dev.Key != 0xffffffff;

         if (p < buf.Length)
            if (MainClass.Opt.Debug)
            Console.WriteLine ("In device info buffer found extra data. Size: {0}",
               buf.Length - p);
         
         return dev;
      }
      
      public string GetUnicodeString (byte[] buf, ref int pos)
      {
         if ((pos + sizeof (int)) > buf.Length)
            return "";
         
         int len = BitConverter.ToInt32 (buf, pos) * 2;
         
         pos += sizeof (int);
         
         if ((pos + len) > buf.Length)
            return "";
         
         string tmp = Encoding.Unicode.GetString (buf, pos, len);
         
         pos += len + 2;
         
         return tmp;
      }

      public string GetAsciiString (byte[] buf, ref int pos)
      {
         if ((pos + sizeof (int)) > buf.Length)
            return "";
         
         int len = BitConverter.ToInt32 (buf, pos);
         
         pos += sizeof (int);
         
         if ((pos + len) > buf.Length || len < 1)
            return "";
         
         int str_len = -1;
         
         for (int i = 0; i < len; i++)
            if (buf [pos + i] == 0) {
               
               str_len = i;
               break;
            }
         
         str_len = str_len == -1 ? len : str_len;
            
         string tmp = Encoding.ASCII.GetString (buf, pos, str_len);
         
         pos += len;
         
         return tmp;
      }
      
      public void WriteBuf (string text, byte[] data, int size)
      {
         Console.WriteLine ("{0}: ", text);
         for (int i = 0; i < size; i++)
            Console.Write ("{0} {1}", data [i].ToString ("X2"), ((i+1) % 16 == 0 ? "\n" : ""));
         Console.WriteLine ();
         
         Console.WriteLine ("{0}: ", text);
         for (int i = 0; i < size; i++)
            Console.Write ("{0} {1}", (char) data [i], ((i+1) % 16 == 0 ? "\n" : ""));
         Console.WriteLine ();
      }
   }
}
